//! Metadata around X clients and manipulating them
use crate::data_types::{Border, ColorScheme, WinId};
use crate::xconnection::XConn;

/**
 * Meta-data around a client window that we are handling.
 * Primarily state flags and information used when determining which clients
 * to show for a given monitor and how they are tiled.
 */
#[derive(Debug, PartialEq, Clone)]
pub struct Client {
    id: WinId,
    wm_class: String,
    workspace: usize,
    // state flags
    floating: bool,
    fullscreen: bool,
}

impl Client {
    pub fn new(id: WinId, wm_class: String, workspace: usize, floating: bool) -> Client {
        Client {
            id,
            wm_class,
            workspace,
            floating: floating,
            fullscreen: false,
        }
    }

    pub fn id(&self) -> WinId {
        self.id
    }

    /// The current workspace index that this client is showing on
    pub fn workspace(&self) -> usize {
        self.workspace
    }

    pub fn class(&self) -> &str {
        &self.wm_class
    }

    pub fn focus(&mut self, conn: &dyn XConn, scheme: &ColorScheme) {
        conn.focus_client(self.id);
        self.set_window_border(conn, Border::Focused, scheme);
    }

    pub fn unfocus(&mut self, conn: &dyn XConn, scheme: &ColorScheme) {
        self.set_window_border(conn, Border::Unfocused, scheme);
    }

    fn set_window_border(&mut self, conn: &dyn XConn, border: Border, scheme: &ColorScheme) {
        let color = match border {
            Border::Urgent => scheme.urgent,
            Border::Focused => scheme.highlight,
            Border::Unfocused => scheme.fg_1,
        };

        conn.set_client_border_color(self.id, color);
    }
}
